import pytest
from hypothesis import given
from pandas.testing import assert_series_equal

from .strategies import (
    hyp_settings,
    st_line_array,
    st_multiline_array,
    st_multipoint_array,
    st_multipolygon_array,
    st_point_array,
    st_polygon_array,
    st_ring_array,
)
from spatialpandas import GeoSeries


@given(st_point_array(geoseries=True))
@hyp_settings
def test_point_array_to_geopandas(gp_point):
    result = GeoSeries(gp_point, dtype='point').to_geopandas()
    assert_series_equal(result, gp_point)


@given(st_multipoint_array(geoseries=True))
@hyp_settings
def test_multipoint_array_to_geopandas(gp_multipoint):
    result = GeoSeries(gp_multipoint, dtype='multipoint').to_geopandas()
    assert_series_equal(result, gp_multipoint)


@given(st_line_array(geoseries=True))
@hyp_settings
def test_line_array_to_geopandas(gp_line):
    result = GeoSeries(gp_line, dtype='line').to_geopandas()
    assert_series_equal(result, gp_line)


@given(st_ring_array(geoseries=True))
@hyp_settings
def test_ring_array_to_geopandas(gp_ring):
    result = GeoSeries(gp_ring, dtype='ring').to_geopandas()
    assert_series_equal(result, gp_ring)


@given(st_multiline_array(geoseries=True))
@hyp_settings
def test_multiline_array_to_geopandas(gp_multiline):
    result = GeoSeries(gp_multiline, dtype='multiline').to_geopandas()
    assert_series_equal(result, gp_multiline)


@pytest.mark.slow
@given(st_polygon_array(geoseries=True))
@hyp_settings
def test_polygon_array_to_geopandas(gp_polygon):
    result = GeoSeries(gp_polygon, dtype='polygon').to_geopandas()
    assert_series_equal(result, gp_polygon)


@pytest.mark.slow
@given(st_multipolygon_array(geoseries=True))
@hyp_settings
def test_multipolygon_array_to_geopandas(gp_multipolygon):
    result = GeoSeries(gp_multipolygon, dtype='multipolygon').to_geopandas()
    assert_series_equal(result, gp_multipolygon)



from functools import total_ordering

import numpy as np
import pyarrow as pa

from ..geometry.base import Geometry, GeometryArray, GeometryDtype
from ..geometry.baselist import _lexographic_lt
from ..utils import _asarray_maybe_ragged
from ._algorithms.bounds import (bounds_interleaved, total_bounds_interleaved,
                                 total_bounds_interleaved_1d)


@total_ordering
class GeometryFixed(Geometry):
    """
    Base class for elements of GeometryFixedArray subclasses
    """
    def __init__(self, data, dtype=None):
        if isinstance(data, np.ndarray):
            # Convert numpy array to bytes
            dtype = data.dtype
            data = data.tobytes()

        super().__init__(data)
        self.numpy_dtype = np.dtype(dtype)
        self.pyarrow_type = pa.from_numpy_dtype(dtype)

    def __lt__(self, other):
        return _lexographic_lt(self.flat_values, other.flat_values)

    def __len__(self):
        return 8 * len(self.data.as_py()) // self.pyarrow_type.bit_width

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.flat_values.tolist())

    @property
    def flat_values(self):
        return pa.Array.from_buffers(
            self.pyarrow_type,
            len(self),
            buffers=[None, pa.py_buffer(self.data.as_py())]
        ).to_numpy()

    @classmethod
    def construct_array_type(cls):
        return GeometryFixedArray


class GeometryFixedArray(GeometryArray):
    """
    Base class for geometry arrays that are backed by a pyarrow fixed width binary type
    """
    _element_type = GeometryFixed
    _element_len = 2

    @classmethod
    def _arrow_type_from_numpy_element_dtype(cls, dtype):
        # Scalar element dtype
        arrow_dtype = pa.from_numpy_dtype(dtype)
        return pa.binary(arrow_dtype.bit_width // 8)

    def _numpy_element_dtype_from_arrow_type(self, pyarrow_type):
        return self._numpy_dtype

    # Constructor
    def __init__(self, array, dtype=None):

        def invalid_array():
            err_msg = (
                "Invalid array with type {typ}\n"
                "A {cls} may be constructed from:\n"
                "    - A 1-d array with length divisible by {n} of interleaved\n"
                "      x y coordinates\n"
                "    - A tuple of {n} 1-d arrays\n"
                "    - A pyarrow.FixedSizeBinaryArray. In this case the dtype\n"
                "      argument must also be specified"
            ).format(
                typ=type(array), cls=self.__class__.__name__, n=self._element_len,
            )
            raise ValueError(err_msg)

        if isinstance(dtype, GeometryDtype):
            dtype = dtype.subtype

        numpy_dtype = None
        pa_type = None
        if isinstance(array, (pa.Array, pa.ChunkedArray)):
            if dtype is None:
                invalid_array()
            numpy_dtype = np.dtype(dtype)
        elif isinstance(array, tuple):
            if len(array) == self._element_len:
                array = [np.asarray(array[i]) for i in range(len(array))]
                if dtype:
                    array = [array[i].astype(dtype) for i in range(len(array))]

                # Capture numpy dtype
                numpy_dtype = array[0].dtype

                # Create buffer and FixedSizeBinaryArray
                pa_type = pa.binary(
                    self._element_len * pa.from_numpy_dtype(numpy_dtype).bit_width // 8
                )
                buffers = [None, pa.py_buffer(np.stack(array, axis=1).tobytes())]
                array = pa.Array.from_buffers(
                    pa_type, len(array[0]), buffers=buffers
                )
            else:
                invalid_array()
        else:
            array = _asarray_maybe_ragged(array)
            if array.dtype.kind == 'O':
                if array.ndim != 1:
                    invalid_array()

                # Try to infer dtype
                if dtype is None:
                    for i in range(len(array)):
                        el = array[i]
                        if el is None:
                            continue
                        if isinstance(el, GeometryFixed):
                            numpy_dtype = el.numpy_dtype
                        else:
                            el_array = np.asarray(el)
                            numpy_dtype = el_array.dtype
                        break
                    if numpy_dtype is None:
                        invalid_array()
                else:
                    numpy_dtype = dtype

                # Explicitly set the pyarrow binary type
                pa_type = pa.binary(
                    self._element_len * pa.from_numpy_dtype(numpy_dtype).bit_width // 8
                )

                # Convert individual rows to bytes
                array = array.copy()
                for i in range(len(array)):
                    el = array[i]
                    if el is None:
                        continue
                    if isinstance(el, bytes):
                        # Nothing to do
                        pass
                    elif isinstance(el, GeometryFixed):
                        array[i] = el.flat_values.tobytes()
                    elif isinstance(el, list) or isinstance(el, np.ndarray):
                        array[i] = np.asarray(el, dtype=numpy_dtype).tobytes()
                    else:
                        array[i] = np.asarray(el.coords, dtype=numpy_dtype).tobytes()
            else:
                if dtype:
                    array = array.astype(dtype)

                # Capture numpy dtype
                numpy_dtype = array.dtype
                pa_type = pa.binary(
                    self._element_len * pa.from_numpy_dtype(numpy_dtype).bit_width // 8
                )

                if array.ndim == 2:
                    # Handle 2d array case
                    if array.shape[1] != self._element_len:
                        invalid_array()
                    # Create buffer and FixedSizeBinaryArray
                    buffers = [None, pa.py_buffer(array.tobytes())]
                    array = pa.Array.from_buffers(
                        pa_type, array.shape[0], buffers=buffers
                    )
                elif array.ndim == 1 and len(array) % self._element_len == 0:
                    buffers = [None, pa.py_buffer(array.tobytes())]
                    array = pa.Array.from_buffers(
                        pa_type, len(array) // self._element_len, buffers=buffers
                    )
                else:
                    invalid_array()

        self._numpy_dtype = numpy_dtype
        super().__init__(array, pa_type)

    # Base geometry methods
    @property
    def flat_values(self):
        if len(self.data) == 0:
            return np.array([], dtype=self.numpy_dtype)
        else:
            start = self.data.offset * self._element_len
            stop = start + len(self.data) * self._element_len
            return np.asarray(self.data.buffers()[1]).view(self.numpy_dtype)[start:stop]

    @property
    def total_bounds(self):
        return total_bounds_interleaved(self.flat_values)

    @property
    def total_bounds_x(self):
        return total_bounds_interleaved_1d(self.flat_values, 0)

    @property
    def total_bounds_y(self):
        return total_bounds_interleaved_1d(self.flat_values, 1)

    @property
    def bounds(self):
        flat_values = self.flat_values
        if len(flat_values) == 0:
            return np.zeros((0, 4), dtype=self.numpy_dtype)
        else:
            bounds = np.full((len(self), 4), np.nan, dtype=np.float64)
            bounds[~self.isna(), :] = bounds_interleaved(
                flat_values, np.arange(0, len(flat_values) + 1, self._element_len)
            )
            return bounds



import numpy as np

from ..utils import ngjit

"""
Initially based on https://github.com/galtay/hilbert_curve, but specialized
for 2 dimensions with numba acceleration
"""


@ngjit
def _int_2_binary(v, width):
    """Return a binary byte array representation of `v` zero padded to `width`
    bits."""
    res = np.zeros(width, dtype=np.uint8)
    for i in range(width):
        res[width - i - 1] = v % 2
        v = v >> 1
    return res


@ngjit
def _binary_2_int(bin_vec):
    """Convert a binary byte array to an integer"""
    res = 0
    next_val = 1
    width = len(bin_vec)
    for i in range(width):
        res += next_val*bin_vec[width - i - 1]
        next_val <<= 1
    return res


@ngjit
def _hilbert_integer_to_transpose(p, h, n):
    """Store a hilbert integer (`h`) as its transpose (`x`).

    Args:
        p (int): iterations to use in the hilbert curve
        h (int): integer distance along hilbert curve
        n (int): number of dimensions
    Returns:
        x (list): transpose of h
                  (n components with values between 0 and 2**p-1)
    """
    h_bits = _int_2_binary(h, p * n)

    x = [_binary_2_int(h_bits[i::n]) for i in range(n)]
    return x


@ngjit
def _transpose_to_hilbert_integer(p, coord):
    """Restore a hilbert integer (`h`) from its transpose (`x`).

    Args:
        p (int): iterations to use in the hilbert curve
        x (list): transpose of h
                  (n components with values between 0 and 2**p-1)

    Returns:
        h (int): integer distance along hilbert curve
    """
    n = len(coord)
    bins = [_int_2_binary(v, p) for v in coord]
    concat = np.zeros(n*p, dtype=np.uint8)
    for i in range(p):
        for j in range(n):
            concat[n*i + j] = bins[j][i]

    h = _binary_2_int(concat)
    return h


@ngjit
def coordinate_from_distance(p, n, h):
    """Return the coordinate for a hilbert distance.

    Args:
        p (int): iterations to use in the hilbert curve
        n (int): number of dimensions
        h (int): integer distance along hilbert curve
    Returns:
        coord (list): Coordinate as length-n list
    """
    coord = _hilbert_integer_to_transpose(p, h, n)
    Z = 2 << (p-1)

    # Gray decode by H ^ (H/2)
    t = coord[n-1] >> 1
    for i in range(n-1, 0, -1):
        coord[i] ^= coord[i-1]
    coord[0] ^= t

    # Undo excess work
    Q = 2
    while Q != Z:
        P = Q - 1
        for i in range(n-1, -1, -1):
            if coord[i] & Q:
                # invert
                coord[0] ^= P
            else:
                # exchange
                t = (coord[0] ^ coord[i]) & P
                coord[0] ^= t
                coord[i] ^= t
        Q <<= 1

    return coord


@ngjit
def coordinates_from_distances(p, n, h):
    """Return the coordinates for an array of hilbert distances.

    Args:
        p (int): iterations to use in the hilbert curve
        n (int): number of dimensions
        h (ndarray): 1d array of integer distances along hilbert curve
    Returns:
        coords (list): 2d array of coordinate, each row a coordinate corresponding to
            associated distance value in input.
    """
    result = np.zeros((len(h), n), dtype=np.int64)

    for i in range(len(h)):
        result[i, :] = coordinate_from_distance(p, n, h[i])

    return result


@ngjit
def distance_from_coordinate(p, coord):
    """Return the hilbert distance for a given coordinate.

    Args:
        p (int): iterations to use in the hilbert curve
        coords (ndarray): coordinate as 1d array
    Returns:
        h (int): distance
    """
    n = len(coord)
    M = 1 << (p - 1)
    # Inverse undo excess work
    Q = M
    while Q > 1:
        P = Q - 1
        for i in range(n):
            if coord[i] & Q:
                coord[0] ^= P
            else:
                t = (coord[0] ^ coord[i]) & P
                coord[0] ^= t
                coord[i] ^= t
        Q >>= 1
    # Gray encode
    for i in range(1, n):
        coord[i] ^= coord[i - 1]
    t = 0
    Q = M
    while Q > 1:
        if coord[n - 1] & Q:
            t ^= Q - 1
        Q >>= 1
    for i in range(n):
        coord[i] ^= t
    h = _transpose_to_hilbert_integer(p, coord)
    return h


@ngjit
def distances_from_coordinates(p, coords):
    """Return the hilbert distances for a given set of coordinates.

    Args:
        p (int): iterations to use in the hilbert curve
        coords (ndarray): 2d array of coordinates, one coordinate per row
    Returns:
        h (ndarray): 1d array of distances
    """
    coords = np.atleast_2d(coords).copy()
    result = np.zeros(coords.shape[0], dtype=np.int64)
    for i in range(coords.shape[0]):
        coord = coords[i, :]
        result[i] = distance_from_coordinate(p, coord)
    return result



from .polygon import Polygon, PolygonArray, PolygonDtype  # noqa
from .multipolygon import ( # noqa
    MultiPolygon, MultiPolygonArray, MultiPolygonDtype
)
from .line import Line, LineArray, LineDtype  # noqa
from .multiline import ( # noqa
    MultiLine, MultiLineArray, MultiLineDtype
)
from .multipoint import ( # noqa
    MultiPoint, MultiPointArray, MultiPointDtype
)
from .ring import Ring, RingArray, RingDtype  # noqa
from .point import Point, PointArray, PointDtype # noqa
from .base import ( # noqa
    Geometry, GeometryArray, GeometryDtype, to_geometry_array
)



import json
from functools import reduce
from glob import has_magic
from numbers import Number
from packaging.version import Version
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import fsspec
import pandas as pd
import pyarrow as pa
from dask import delayed
from dask.dataframe import from_delayed, from_pandas
from dask.dataframe import read_parquet as dd_read_parquet
from dask.dataframe import to_parquet as dd_to_parquet  # noqa
from dask.utils import natural_sort_key
from pandas.io.parquet import to_parquet as pd_to_parquet
from pyarrow.parquet import ParquetDataset, ParquetFile, read_metadata

from .. import GeoDataFrame
from ..dask import DaskGeoDataFrame
from ..geometry import GeometryDtype
from ..io.utils import (
    PathType,
    _maybe_prepend_protocol,
    validate_coerce_filesystem,
)

# improve pandas compatibility, based on geopandas _compat.py
PANDAS_GE_12 = Version(pd.__version__) >= Version("1.2.0")

# When we drop support for pyarrow < 5 all code related to this can be removed.
LEGACY_PYARROW = Version(pa.__version__) < Version("5.0.0")


def _load_parquet_pandas_metadata(
    path,
    filesystem=None,
    storage_options=None,
    engine_kwargs=None,
):
    engine_kwargs = engine_kwargs or {}
    filesystem = validate_coerce_filesystem(path, filesystem, storage_options)
    if not filesystem.exists(path):
        raise ValueError("Path not found: " + path)

    if filesystem.isdir(path):
        if LEGACY_PYARROW:
            basic_kwargs = dict(validate_schema=False)
        else:
            basic_kwargs = dict(use_legacy_dataset=False)

        pqds = ParquetDataset(
            path,
            filesystem=filesystem,
            **basic_kwargs,
            **engine_kwargs,
        )

        if LEGACY_PYARROW:
            common_metadata = pqds.common_metadata
            if common_metadata is None:
                # Get metadata for first piece
                piece = pqds.pieces[0]
                metadata = piece.get_metadata().metadata
            else:
                metadata = pqds.common_metadata.metadata
        else:
            filename = "/".join([_get_parent_path(pqds.files[0]), "_common_metadata"])
            try:
                common_metadata = _read_metadata(filename, filesystem=filesystem)
            except FileNotFoundError:
                # Common metadata doesn't exist, so get metadata for first piece instead
                filename = pqds.files[0]
                common_metadata = _read_metadata(filename, filesystem=filesystem)
            metadata = common_metadata.metadata
    else:
        with filesystem.open(path) as f:
            pf = ParquetFile(f)
        metadata = pf.metadata.metadata

    return json.loads(
        metadata.get(b'pandas', b'{}').decode('utf')
    )


def to_parquet(
    df: GeoDataFrame,
    path: PathType,
    compression: Optional[str] = "snappy",
    filesystem: Optional[fsspec.AbstractFileSystem] = None,
    index: Optional[bool] = None,
    storage_options: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> None:
    if filesystem is not None:
        filesystem = validate_coerce_filesystem(path, filesystem, storage_options)

    # Standard pandas to_parquet with pyarrow engine
    to_parquet_args = {
        "df": df,
        "path": path,
        "engine": "pyarrow",
        "compression": compression,
        "filesystem": filesystem,
        "index": index,
        **kwargs,
    }

    if PANDAS_GE_12:
        to_parquet_args.update({"storage_options": storage_options})
    else:
        if filesystem is None:
            filesystem = validate_coerce_filesystem(path, filesystem, storage_options)
            to_parquet_args.update({"filesystem": filesystem})

    pd_to_parquet(**to_parquet_args)


def read_parquet(
    path: PathType,
    columns: Optional[Iterable[str]] = None,
    filesystem: Optional[fsspec.AbstractFileSystem] = None,
    storage_options: Optional[Dict[str, Any]] = None,
    engine_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> GeoDataFrame:
    engine_kwargs = engine_kwargs or {}
    filesystem = validate_coerce_filesystem(path, filesystem, storage_options)

    if LEGACY_PYARROW:
        basic_kwargs = dict(validate_schema=False)
    else:
        basic_kwargs = dict(use_legacy_dataset=False)

    # Load using pyarrow to handle parquet files and directories across filesystems
    dataset = ParquetDataset(
        path,
        filesystem=filesystem,
        **basic_kwargs,
        **engine_kwargs,
        **kwargs,
    )

    if LEGACY_PYARROW:
        metadata = _load_parquet_pandas_metadata(
            path,
            filesystem=filesystem,
            storage_options=storage_options,
            engine_kwargs=engine_kwargs,
        )
    else:
        metadata = dataset.schema.pandas_metadata

    # If columns specified, prepend index columns to it
    if columns is not None:
        all_columns = set(column['name'] for column in metadata.get('columns', []))
        index_col_metadata = metadata.get('index_columns', [])
        extra_index_columns = []
        for idx_metadata in index_col_metadata:
            if isinstance(idx_metadata, str):
                name = idx_metadata
            elif isinstance(idx_metadata, dict):
                name = idx_metadata.get('name', None)
            else:
                name = None
            if name is not None and name not in columns and name in all_columns:
                extra_index_columns.append(name)

        columns = extra_index_columns + list(columns)

    df = dataset.read(columns=columns).to_pandas()

    # Return result
    return GeoDataFrame(df)


def to_parquet_dask(
    ddf: DaskGeoDataFrame,
    path: PathType,
    compression: Optional[str] = "snappy",
    filesystem: Optional[fsspec.AbstractFileSystem] = None,
    storage_options: Optional[Dict[str, Any]] = None,
    engine_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> None:
    engine_kwargs = engine_kwargs or {}

    if not isinstance(ddf, DaskGeoDataFrame):
        raise TypeError(f"Expected DaskGeoDataFrame not {type(ddf)}")
    filesystem = validate_coerce_filesystem(path, filesystem, storage_options)
    if kwargs.get("overwrite", False) and \
        not kwargs.get("append", False) and \
            path and filesystem.isdir(path):
        filesystem.rm(path, recursive=True)

    # Determine partition bounding boxes to save to _metadata file
    partition_bounds = {}
    for series_name in ddf.columns:
        series = ddf[series_name]
        if isinstance(series.dtype, GeometryDtype):
            partition_bounds[series_name] = series.partition_bounds.to_dict()

    spatial_metadata = {'partition_bounds': partition_bounds}
    b_spatial_metadata = json.dumps(spatial_metadata).encode('utf')

    dd_to_parquet(
        ddf,
        path,
        engine="pyarrow",
        compression=compression,
        storage_options=storage_options,
        custom_metadata={b'spatialpandas': b_spatial_metadata},
        write_metadata_file=True,
        **engine_kwargs,
        **kwargs,
    )


def read_parquet_dask(
    path: PathType,
    columns: Optional[Iterable[str]] = None,
    filesystem: Optional[fsspec.AbstractFileSystem] = None,
    load_divisions: Optional[bool] = False,
    geometry: Optional[str] = None,
    bounds: Optional[Tuple[Number, Number, Number, Number]] = None,
    categories: Optional[Union[List[str], Dict[str, str]]] = None,
    build_sindex: Optional[bool] = False,
    storage_options: Optional[Dict[str, Any]] = None,
    engine_kwargs: Optional[Dict[str, Any]] = None,
) -> DaskGeoDataFrame:
    """Read spatialpandas parquet dataset(s) as DaskGeoDataFrame.

    Datasets are assumed to have been written with the
    DaskGeoDataFrame.to_parquet or
    DaskGeoDataFrame.pack_partitions_to_parquet methods.

    Args:
        path: Path to spatialpandas parquet dataset, or list of paths to datasets, or
            glob string referencing one or more parquet datasets.
        columns: List of columns to load
        filesystem: fsspec filesystem to use to read datasets
        load_divisions: If True, attempt to load the hilbert_distance divisions for
            each partition.  Only available for datasets written using the
            pack_partitions_to_parquet method.
        geometry: The name of the column to use as the geometry column of the
            resulting DaskGeoDataFrame. Defaults to the first geometry column in the
            dataset.
        bounds: If specified, only load partitions that have the potential to intersect
            with the provided bounding box coordinates. bounds is a length-4 tuple of
            the form (xmin, ymin, xmax, ymax).
        categories : list, dict or None
            For any fields listed here, if the parquet encoding is Dictionary,
            the column will be created with dtype category. Use only if it is
            guaranteed that the column is encoded as dictionary in all row-groups.
            If a list, assumes up to 2**16-1 labels; if a dict, specify the number
            of labels expected; if None, will load categories automatically for
            data written by dask/fastparquet, not otherwise.
        build_sindex : boolean
            Whether to build partition level spatial indexes to speed up indexing.
        storage_options: Key/value pairs to be passed on to the file-system backend, if any.
        engine_kwargs: pyarrow.parquet engine-related keyword arguments.
    Returns:
    DaskGeoDataFrame
    """
    # Normalize path to a list
    if isinstance(path, (str, Path)):
        paths = [path]
    else:
        paths = list(path)
        if not paths:
            raise ValueError('Empty path specification')

    # Infer filesystem
    filesystem = validate_coerce_filesystem(
        paths[0],
        filesystem,
        storage_options,
    )

    files = []
    for p in paths:
        if hasattr(p, "as_posix"):
            p = p.as_posix()
        if has_magic(p):
            # Expand glob paths
            _files = _expand_path(p, filesystem)
            files.extend(_files)
        elif not filesystem.exists(p):
            raise FileNotFoundError(p)
        else:
            files.append(p)

    # Perform read parquet
    result = _perform_read_parquet_dask(
        files,
        columns=columns,
        filesystem=filesystem,
        load_divisions=load_divisions,
        geometry=geometry,
        bounds=bounds,
        categories=categories,
        storage_options=storage_options,
        engine_kwargs=engine_kwargs,
    )

    if build_sindex:
        result = result.build_sindex()

    return result


def _expand_path(paths, filesystem):
    # Expand glob paths
    files = filesystem.expand_path(paths)
    if isinstance(files, str):
        files = [files]
    # Filter out metadata files
    files = [_ for _ in files if "_metadata" not in _.split("/")[-1]]
    files = _maybe_prepend_protocol(files, filesystem)
    return files


def _perform_read_parquet_dask(
    paths,
    columns,
    filesystem,
    load_divisions,
    geometry=None,
    bounds=None,
    categories=None,
    storage_options=None,
    engine_kwargs=None,
):
    engine_kwargs = engine_kwargs or {}
    filesystem = validate_coerce_filesystem(
        paths[0],
        filesystem,
        storage_options,
    )
    if LEGACY_PYARROW:
        basic_kwargs = dict(validate_schema=False)
    else:
        basic_kwargs = dict(use_legacy_dataset=False)

    datasets = []
    for path in paths:
        if filesystem.isdir(path):
            path = path.rstrip('/')
            path = f"{path}/**/*.parquet"
        if has_magic(path):
            path = _expand_path(path, filesystem)
        d = ParquetDataset(
            path,
            filesystem=filesystem,
            **basic_kwargs,
            **engine_kwargs,
        )
        datasets.append(d)

    # Create delayed partition for each piece
    pieces = []
    for dataset in datasets:
        # Perform natural sort on pieces so that "part.10" comes after "part.2"
        fragments = getattr(dataset, "fragments", None)
        if fragments is None:
            fragments = dataset.pieces
        dataset_pieces = sorted(fragments, key=lambda piece: natural_sort_key(piece.path))
        pieces.extend(dataset_pieces)

    delayed_partitions = [
        delayed(read_parquet)(
            piece.path,
            columns=columns,
            filesystem=filesystem,
            storage_options=storage_options,
            engine_kwargs=engine_kwargs,
        )
        for piece in pieces
    ]

    # Load divisions
    if load_divisions:
        div_mins_list, div_maxes_list = zip(*[
            _load_divisions(dataset) for dataset in datasets
        ])

        div_mins = reduce(lambda a, b: a + b, div_mins_list, [])
        div_maxes = reduce(lambda a, b: a + b, div_maxes_list, [])
    else:
        div_mins = None
        div_maxes = None

    # load partition bounds
    partition_bounds_list = [
        _load_partition_bounds(dataset, filesystem=filesystem)
        for dataset in datasets
    ]
    if not any([b is None for b in partition_bounds_list]):
        partition_bounds = {}
        # We have partition bounds for all datasets
        for partition_bounds_el in partition_bounds_list:
            for col, col_bounds in partition_bounds_el.items():
                col_bounds_list = partition_bounds.get(col, [])
                col_bounds_list.append(col_bounds)
                partition_bounds[col] = col_bounds_list

        # Concat bounds for each geometry column
        for col in list(partition_bounds):
            partition_bounds[col] = pd.concat(
                partition_bounds[col], axis=0
            ).reset_index(drop=True)
            partition_bounds[col].index.name = 'partition'
    else:
        partition_bounds = {}

    # Use Dask's read_parquet to get metadata
    if columns is not None:
        cols_no_index = [col for col in columns if col != "hilbert_distance"]
    else:
        cols_no_index = None

    if LEGACY_PYARROW:
        files = paths
    else:
        files = getattr(datasets[0], "files", paths)

    meta = dd_read_parquet(
        files[0],
        columns=cols_no_index,
        filesystem=filesystem,
        engine='pyarrow',
        categories=categories,
        ignore_metadata_file=True,
        **engine_kwargs,
    )._meta

    meta = GeoDataFrame(meta)

    # Handle geometry in meta
    if geometry:
        meta = meta.set_geometry(geometry)

    geometry = meta.geometry.name

    # Filter partitions by bounding box
    if bounds and geometry in partition_bounds:
        # Unpack bounds coordinates and make sure
        x0, y0, x1, y1 = bounds

        # Make sure x0 < c1
        if x0 > x1:
            x0, x1 = x1, x0

        # Make sure y0 < y1
        if y0 > y1:
            y0, y1 = y1, y0

        # Make DataFrame with bounds and parquet piece
        partitions_df = partition_bounds[geometry].assign(
            delayed_partition=delayed_partitions
        )

        if load_divisions:
            partitions_df = partitions_df.assign(div_mins=div_mins, div_maxes=div_maxes)

        inds = ~(
            (partitions_df.x1 < x0) |
            (partitions_df.y1 < y0) |
            (partitions_df.x0 > x1) |
            (partitions_df.y0 > y1)
        )

        partitions_df = partitions_df[inds]
        for col in list(partition_bounds):
            partition_bounds[col] = partition_bounds[col][inds]
            partition_bounds[col].reset_index(drop=True, inplace=True)
            partition_bounds[col].index.name = "partition"

        delayed_partitions = partitions_df.delayed_partition.tolist()
        if load_divisions:
            div_mins = partitions_df.div_mins
            div_maxes = partitions_df.div_maxes

    if load_divisions:
        divisions = div_mins + [div_maxes[-1]]
        if divisions != sorted(divisions):
            raise ValueError(
                "Cannot load divisions because the discovered divisions are unsorted.\n"
                "Set load_divisions=False to skip loading divisions."
            )
    else:
        divisions = None

    # Create DaskGeoDataFrame
    if delayed_partitions:
        result = from_delayed(
            delayed_partitions, divisions=divisions, meta=meta, verify_meta=False
        )
    else:
        # Single partition empty result
        result = from_pandas(meta, npartitions=1)

    # Set partition bounds
    if partition_bounds:
        result._partition_bounds = partition_bounds

    return result


def _get_parent_path(path):
    if isinstance(path, Iterable) and not isinstance(path, str):
        path = path[0]
    parent = str(path).rsplit("/", 1)[0]
    return parent if parent else "/"


def _read_metadata(filename, filesystem):
    with filesystem.open(filename, "rb") as f:
        common_metadata = read_metadata(f)
    return common_metadata


def _load_partition_bounds(pqds, filesystem=None):
    partition_bounds = None

    if LEGACY_PYARROW:
        common_metadata = pqds.common_metadata
    else:
        filename = "/".join([_get_parent_path(pqds.files[0]), "_common_metadata"])
        try:
            common_metadata = _read_metadata(filename, filesystem=filesystem)
        except FileNotFoundError:
            common_metadata = None

    if common_metadata is not None and b'spatialpandas' in common_metadata.metadata:
        spatial_metadata = json.loads(
            common_metadata.metadata[b'spatialpandas'].decode('utf')
        )
        if "partition_bounds" in spatial_metadata:
            partition_bounds = {}
            for name in spatial_metadata['partition_bounds']:
                bounds_df = pd.DataFrame(
                    spatial_metadata['partition_bounds'][name]
                )

                # Index labels will be read in as strings.
                # Here we convert to integers, sort by index, then drop index just in
                # case the rows got shuffled on read
                bounds_df = (bounds_df
                             .set_index(bounds_df.index.astype('int'))
                             .sort_index()
                             .reset_index(drop=True))
                bounds_df.index.name = 'partition'

                partition_bounds[name] = bounds_df

    return partition_bounds


def _load_divisions(pqds):
    fmd = pqds.metadata
    row_groups = [fmd.row_group(i) for i in range(fmd.num_row_groups)]
    rg0 = row_groups[0]
    div_col = None
    for c in range(rg0.num_columns):
        if rg0.column(c).path_in_schema == "hilbert_distance":
            div_col = c
            break

    if div_col is None:
        # No hilbert_distance column found
        raise ValueError(
            "Cannot load divisions because hilbert_distance index not found"
        )

    mins, maxes = zip(*[
        (rg.column(div_col).statistics.min, rg.column(div_col).statistics.max)
        for rg in row_groups
    ])

    return list(mins), list(maxes)



from itertools import product

import hypothesis.strategies as st
import numpy as np
import pytest
from hypothesis import given, settings

try:
    from hilbertcurve.hilbertcurve import HilbertCurve
except ImportError:
    HilbertCurve = None

# ### hypothesis settings ###
from spatialpandas.spatialindex.hilbert_curve import (
    coordinate_from_distance,
    coordinates_from_distances,
    distance_from_coordinate,
    distances_from_coordinates,
)

hyp_settings = settings(deadline=None)

# ### strategies ###
st_p = st.integers(min_value=1, max_value=5)
st_n = st.integers(min_value=1, max_value=3)

# ### Hypothesis tests ###
@given(st_p, st_n)
@hyp_settings
@pytest.mark.skipif(HilbertCurve is None, reason="hilbertcurve not installed")
def test_coordinates_from_distance(p, n):
    # Build vector of possible distances
    distances = np.arange(2 ** (n * p), dtype=np.int64)

    # Compute coordinates as vector result
    vector_result = coordinates_from_distances(p, n, distances)

    # Compare with reference
    reference_hc = HilbertCurve(p, n)
    for i, distance in enumerate(distances):
        # Reference
        expected = tuple(reference_hc.point_from_distance(distance))

        # Scalar result
        scalar_result = tuple(coordinate_from_distance(p, n, distance))
        assert scalar_result == expected

        # Vector result
        assert tuple(vector_result[i, :]) == expected


@given(st_p, st_n)
@hyp_settings
@pytest.mark.skipif(HilbertCurve is None, reason="hilbertcurve not installed")
def test_distance_from_coordinates(p, n):
    side_len = 2 ** p

    # build matrix of all possible coordinate
    coords = np.array(list(product(range(side_len), repeat=n)))

    # Compute distances as vector result
    vector_result = distances_from_coordinates(p, coords)

    # Compare with reference
    reference_hc = HilbertCurve(p, n)
    for i in range(coords.shape[0]):
        coord = coords[i, :]

        # Reference
        expected = reference_hc.distance_from_point(coord)

        # Compute scalar distance and compare
        scalar_result = distance_from_coordinate(p, coord)
        assert scalar_result == expected

        # Compare with vector distance compute above
        assert vector_result[i] == expected



import numpy as np
import pytest
from hypothesis import example, given
from shapely import geometry as sg

from ..strategies import (
    coord,
    hyp_settings,
    st_bounds,
    st_line_array,
    st_multiline_array,
    st_multipoint_array,
    st_multipolygon_array,
    st_point_array,
    st_points,
    st_polygon,
    st_polygon_array,
)
from spatialpandas.geometry import (
    LineArray,
    MultiLineArray,
    MultiPointArray,
    MultiPolygonArray,
    PointArray,
    Polygon,
    PolygonArray,
)
from spatialpandas.geometry._algorithms.intersection import (
    point_intersects_polygon,
    segment_intersects_point,
    segments_intersect,
)


@given(coord, coord, coord, coord, coord, coord)
@example(0, 0, 1, 1, 0.25, 0.25)  # Point on line
@example(0, 0, 1, 1, 0.2501, 0.25)  # Point just off of line
@example(0, 0, 1, 1, 1, 1)  # Point on end of segment
@example(0, 0, 1, 1, 0, 0)  # Point on start of segment
@example(0, 0, 1, 1, 1.001, 1.001)  # Point on ray, just past end of segment
@example(0, 0, 1, 1, -0.001, -0.001)  # Point on ray, just before start of segment
@hyp_settings
def test_segment_intersects_point(ax0, ay0, ax1, ay1, bx, by):
    result1 = segment_intersects_point(ax0, ay0, ax1, ay1, bx, by)
    result2 = segment_intersects_point(ax1, ay1, ax0, ay0, bx, by)
    # Ensure we get the same answer with line order flipped
    assert result1 == result2

    # Use shapely polygon to compute expected intersection
    line = sg.LineString([(ax0, ay0), (ax1, ay1)])
    point = sg.Point([bx, by])
    expected = line.intersects(point)
    assert result1 == expected


@given(coord, coord, coord, coord, coord, coord, coord, coord)
@hyp_settings
def test_segment_intersection(ax0, ay0, ax1, ay1, bx0, by0, bx1, by1):
    result1 = segments_intersect(ax0, ay0, ax1, ay1, bx0, by0, bx1, by1)
    result2 = segments_intersect(bx0, by0, bx1, by1, ax0, ay0, ax1, ay1)
    # Ensure we get the same answer with line order flipped
    assert result1 == result2

    # Use shapely polygon to compute expected intersection
    if (ax0 == ax1 == bx0 == bx1) or (ay0 == ay1 == by0 == by1):
        return
    line1 = sg.LineString([(ax0, ay0), (ax1, ay1)])
    line2 = sg.LineString([(bx0, by0), (bx1, by1)])
    expected = line1.intersects(line2)
    assert result1 == expected


@given(st_polygon(), st_points)
@hyp_settings
def test_point_intersects_polygon(sg_polygon, points):
    polygon = Polygon.from_shapely(sg_polygon)

    for r in range(points.shape[0]):
        x, y = points[r, :]
        result = point_intersects_polygon(
            x, y, polygon.buffer_values, polygon.buffer_inner_offsets
        )

        expected = sg_polygon.intersects(sg.Point([x, y]))
        assert expected == result


@given(st_point_array(), st_bounds())
@hyp_settings
def test_point_intersects_rect(gp_point, rect):
    sg_rect = sg.box(*rect)
    expected = gp_point.intersects(sg_rect)
    points = PointArray.from_geopandas(gp_point)

    # Test MultiPointArray.intersects_rect
    result = points.intersects_bounds(rect)
    np.testing.assert_equal(result, expected)

    # Test MultiPointArray.intersects_rect with inds
    inds = np.flipud(np.arange(0, len(points)))
    result = points.intersects_bounds(rect, inds)
    np.testing.assert_equal(result, np.flipud(expected))

    # Test MultiPoint.intersects_rect
    result = np.array([
        point.intersects_bounds(rect) for point in points
    ])
    np.testing.assert_equal(result, expected)


@given(st_multipoint_array(), st_bounds())
@hyp_settings
def test_multipoint_intersects_rect(gp_multipoint, rect):
    sg_rect = sg.box(*rect)
    expected = gp_multipoint.intersects(sg_rect)
    multipoints = MultiPointArray.from_geopandas(gp_multipoint)

    # Test MultiPointArray.intersects_rect
    result = multipoints.intersects_bounds(rect)
    np.testing.assert_equal(result, expected)

    # Test MultiPointArray.intersects_rect with inds
    inds = np.flipud(np.arange(0, len(multipoints)))
    result = multipoints.intersects_bounds(rect, inds)
    np.testing.assert_equal(result, np.flipud(expected))

    # Test MultiPoint.intersects_rect
    result = np.array([
        multipoint.intersects_bounds(rect) for multipoint in multipoints
    ])
    np.testing.assert_equal(result, expected)


@given(st_line_array(), st_bounds())
@hyp_settings
def test_line_intersects_rect(gp_line, rect):
    x0, y0, x1, y1 = rect
    if x0 == x1 or y0 == y1:
        return
    sg_rect = sg.box(*rect)

    expected = gp_line.intersects(sg_rect)
    lines = LineArray.from_geopandas(gp_line)

    # Test LineArray.intersects_rect
    result = lines.intersects_bounds(rect)
    np.testing.assert_equal(result, expected)

    # Test LineArray.intersects_rect with inds
    inds = np.flipud(np.arange(0, len(lines)))
    result = lines.intersects_bounds(rect, inds)
    np.testing.assert_equal(result, np.flipud(expected))

    # Test Line.intersects_rect
    result = np.array([line.intersects_bounds(rect) for line in lines])
    np.testing.assert_equal(result, expected)


@given(st_multiline_array(), st_bounds())
@hyp_settings
def test_multiline_intersects_rect(gp_multiline, rect):
    x0, y0, x1, y1 = rect
    if x0 == x1 or y0 == y1:
        return
    sg_rect = sg.box(*rect)

    expected = gp_multiline.intersects(sg_rect)
    multilines = MultiLineArray.from_geopandas(gp_multiline)

    # Test MultiLineArray.intersects_rect
    result = multilines.intersects_bounds(rect)
    if not np.array_equal(result, expected):
        multilines.intersects_bounds(rect)
    np.testing.assert_equal(result, expected)

    # Test MultiLineArray.intersects_rect with inds
    inds = np.flipud(np.arange(0, len(multilines)))
    result = multilines.intersects_bounds(rect, inds)
    np.testing.assert_equal(result, np.flipud(expected))

    # Test MultiLine.intersects_rect
    result = np.array([
        multiline.intersects_bounds(rect) for multiline in multilines
    ])
    np.testing.assert_equal(result, expected)


@pytest.mark.slow
@given(
    st_polygon_array(),
    st_bounds(
        x_min=-150, x_max=150, y_min=-150, y_max=150
    )
)
@hyp_settings
def test_polygon_intersects_rect(gp_polygon, rect):
    x0, y0, x1, y1 = rect
    if x0 == x1 or y0 == y1:
        return
    sg_rect = sg.box(*rect)

    expected = gp_polygon.intersects(sg_rect)
    polygons = PolygonArray.from_geopandas(gp_polygon)

    # Test PolygonArray.intersects_rect
    result = polygons.intersects_bounds(rect)
    np.testing.assert_equal(result, expected)

    # Test PolygonArray.intersects_rect with inds
    inds = np.flipud(np.arange(0, len(polygons)))
    result = polygons.intersects_bounds(rect, inds)
    np.testing.assert_equal(result, np.flipud(expected))

    # Test Polygon.intersects_rect
    result = np.array([
        polygon.intersects_bounds(rect) for polygon in polygons
    ])
    np.testing.assert_equal(result, expected)


@pytest.mark.slow
@given(
    st_multipolygon_array(),
    st_bounds(
        x_min=-150, x_max=150, y_min=-150, y_max=150
    )
)
@hyp_settings
def test_multipolygon_intersects_rect(gp_multipolygon, rect):
    x0, y0, x1, y1 = rect
    if x0 == x1 or y0 == y1:
        return
    sg_rect = sg.box(*rect)

    expected = gp_multipolygon.intersects(sg_rect)
    multipolygons = MultiPolygonArray.from_geopandas(gp_multipolygon)

    # Test MultiPolygonArray.intersects_rect
    result = multipolygons.intersects_bounds(rect)
    np.testing.assert_equal(result, expected)

    # Test MultiPolygonArray.intersects_rect with inds
    inds = np.flipud(np.arange(0, len(multipolygons)))
    result = multipolygons.intersects_bounds(rect, inds)
    np.testing.assert_equal(result, np.flipud(expected))

    # Test MultiPolygon.intersects_rect
    result = np.array([
        multipolygon.intersects_bounds(rect) for multipolygon in multipolygons
    ])
    np.testing.assert_equal(result, expected)



import dask.dataframe as dd
import geopandas as gp
import numpy as np
import pandas as pd
import pytest
from hypothesis import given

import spatialpandas as sp
from ..geometry.strategies import st_point_array, st_polygon_array
from ..test_parquet import hyp_settings
from spatialpandas import GeoDataFrame
from spatialpandas.dask import DaskGeoDataFrame

try:
    from geopandas._compat import HAS_RTREE, USE_PYGEOS
    gpd_spatialindex = USE_PYGEOS or HAS_RTREE
except ImportError:
    try:
        import rtree  # noqa
        gpd_spatialindex = rtree
    except Exception:
        gpd_spatialindex = False

if not gpd_spatialindex:
    pytest.skip('Geopandas spatialindex not available to compare against',
                allow_module_level=True)


@pytest.mark.slow
@pytest.mark.parametrize("how", ["inner", "left", "right"])
@given(
    st_point_array(min_size=1, geoseries=True),
    st_polygon_array(min_size=1, geoseries=True),
)
@hyp_settings
def test_sjoin(how, gp_points, gp_polygons):
    # join witgh geopandas
    left_gpdf = gp.GeoDataFrame({
        'geometry': gp_points,
        'a': np.arange(10, 10 + len(gp_points)),
        'c': np.arange(20, 20 + len(gp_points)),
        'v': np.arange(30, 30 + len(gp_points)),
    }).set_index('a')

    right_gpdf = gp.GeoDataFrame({
        'geometry': gp_polygons,
        'a': np.arange(10, 10 + len(gp_polygons)),
        'b': np.arange(20, 20 + len(gp_polygons)),
        'v': np.arange(30, 30 + len(gp_polygons)),
    }).set_index('b')

    # Generate expected result as geopandas GeoDataFrame
    gp_expected = gp.sjoin(left_gpdf, right_gpdf, how=how)
    gp_expected = gp_expected.rename(columns={"v_x": "v_left", "v_y": "v_right"})
    if how == "right":
        gp_expected.index.name = right_gpdf.index.name
    else:
        gp_expected.index.name = left_gpdf.index.name

    # join with spatialpandas
    left_spdf = GeoDataFrame(left_gpdf)
    right_spdf = GeoDataFrame(right_gpdf)

    result = sp.sjoin(
        left_spdf, right_spdf, how=how
    ).sort_values(['v_left', 'v_right'])
    assert isinstance(result, GeoDataFrame)

    # Check pandas results
    if len(gp_expected) == 0:
        assert len(result) == 0
    else:
        expected = GeoDataFrame(gp_expected).sort_values(['v_left', 'v_right'])
        pd.testing.assert_frame_equal(expected, result)

        # left_df as Dask frame
        left_spddf = dd.from_pandas(left_spdf, npartitions=4)

        if how == "right":
            # Right join not supported when left_df is a Dask DataFrame
            with pytest.raises(ValueError):
                sp.sjoin(left_spddf, right_spdf, how=how)
            return
        else:
            result_ddf = sp.sjoin(
                left_spddf, right_spdf, how=how
            )
        assert isinstance(result_ddf, DaskGeoDataFrame)
        assert result_ddf.npartitions <= 4
        result = result_ddf.compute().sort_values(['v_left', 'v_right'])
        pd.testing.assert_frame_equal(expected, result)
